//
// GameX
// Ship Demo - Photon Class Code
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

#include "shipdemox-photons.hpp"

// Photons - Initialize
// Load Photon image and initialize any other resources or variables we may need for the photons
void Photons::Initialize (void)
{
	_photonImg[0].Load ("Images\\photon.bmp"); // photon image needs no alpha layer because it is drawn additively
}

// Photons - AddPhoton
// Add a photon to the list of particles.
int Photons::AddPhoton (const Vec2f & pos, const Vec2f & vel)
{
	// If maximum reached, replace an existing photon at random, otherwise add new photon at the end
	int new_photon = (_numPhotons == MAX_PHOTONS) ? GameX.GetRand(MAX_PHOTONS) : _numPhotons++;
	_photonList[new_photon].pos = pos;	// Initialize new photon with given or default parameters
	_photonList[new_photon].vel = vel;
	_photonList[new_photon].age = 0;
	_photonList[new_photon].img = 0;
	return new_photon;	// Return the number of the photon just added, so it can be kept track of if desired
}

// Photons - Update
// Move all the Photons in the particle system.
// The Photons are implemented as a very simple physics engine for 
// a particle system. Each particle is advanced according to its velocity.
// Particles out of bounds are rebounded into the view.
// Particles which are too old are deleted.
void Photons::Update (void)
{									// Process each particle (photon) for updating/moving:
	for(int n = 0 ; n < _numPhotons ; n++)
	{
		_photonList[n].age++;					// Increment age of photon, and...
		if(_photonList[n].age > MAX_AGE)		// If photon is too old
		{
			// delete it by putting the last one in its place and decrementing the total number of photons
			_photonList[n--] = _photonList[--_numPhotons];
		}
		else	// If photon is still "young" enough:
		{
			_photonList[n].pos += _photonList[n].vel;	// Move photon by its velocity

			// After moving, rebound photon off view edges if it went out of bounds:
			if(_photonList[n].pos.x < 0                && _photonList[n].vel.x < 0
			|| _photonList[n].pos.x > GameX.GetWidth() && _photonList[n].vel.x > 0)
				_photonList[n].vel.x = -_photonList[n].vel.x;

			if(_photonList[n].pos.y < 0                 && _photonList[n].vel.y < 0
			|| _photonList[n].pos.y > GameX.GetHeight() && _photonList[n].vel.y > 0)
				_photonList[n].vel.y = -_photonList[n].vel.y;
		}
	}
}

// Photons - Draw
// Draw all Photons
void Photons::Draw (void)
{										// Process each particle (photon) for drawing:
	for(int n = 0 ; n < _numPhotons ; n++)
	{
		GameX.SetDrawMode (DRAW_ADD | DRAWOP_CENTERED); // Set to draw additively and centered at position.
		GameX.SetDrawScale (1.25f - (float)_photonList[n].age*0.5f/(float)MAX_AGE); // Scale to be smaller when old
		GameX.DrawImage (&_photonImg[0], _photonList[n].pos); // Draw Photon to screen using GameX.
	}
}

Photons g_photons;				// Create global photons object


