//
// GameX
// Ship Demo - Ship Class Code
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

#include "shipdemox-ship.hpp"
#include "shipdemox-photons.hpp"

// Ship - Initialize
// Initialize all ship images and ship's target image
void Ship::Initialize (const int shipNum)
{
	_shipNum = shipNum;				// Initialize ship number													
	_currImg = SHIP_IMG_FLAT;		// Set current ship image
	_refire = 0;
									// Set initial ship position and angle:
	if(_shipNum == 0) // blue ship
	{
		_pos = Vec2f(GameX.GetWidth()/3, GameX.GetHeight()/2); // 1/3 of the screen away from the left of the screen
		_ang = 0.0f; // facing to the right
		_weight = 2.0f;
	}
	else // red ship
	{
		_pos = Vec2f(GameX.GetWidth()*2/3, GameX.GetHeight()/2); // 1/3 of the screen away from the right of the screen
		_ang = 180.0f; // facing to the left
		_weight = 1.0f;
	}
	
									// Load Sound(s):
	_photonSnd.LoadWav("Sounds\\torpedo.wav");

									// Load Images:
	_sprites.Load("Images\\ships.gif", "Images\\ships_alpha.gif"); // Load the sprite sheet image containing the ships, and a transparency layer from a different image file
	_sprites.SwapColors((ColorSwapType)_shipNum); // change to ship's color

	_targetImg.Load("Images\\target.tif", true); // Load the ship's target image, with transparency from the same image file
	_targetImg.SwapColors((ColorSwapType)_shipNum); // change to ship's color
}

// Ship - Update
// Get user input and move the ship
void Ship::Update (void)
{
	bool pressedLeft, pressedRight, pressedUp, pressedDown, pressedFire; // each is true if pressed, false if not

	if(_shipNum == 0) // Process user input for Blue Ship (num 0)...
	{
		pressedLeft = GameX.IsKeyDown(KEY_LEFT);
		pressedRight = GameX.IsKeyDown(KEY_RIGHT);
		pressedUp = GameX.IsKeyDown(KEY_UP);
		pressedDown = GameX.IsKeyDown(KEY_DOWN);
		pressedFire = GameX.IsKeyDown(KEY_SPACE);
	}
	else // Process user input for Red Ship (num 1)...
	{	
		pressedLeft = GameX.IsKeyDown(KEY_A);
		pressedRight = GameX.IsKeyDown(KEY_D);
		pressedUp = GameX.IsKeyDown(KEY_W);
		pressedDown = GameX.IsKeyDown(KEY_S);
		pressedFire = GameX.IsKeyDown(KEY_X);
	}

	if(pressedLeft && pressedRight) pressedLeft = false, pressedRight = false; // Cancel out any
	if(pressedUp   && pressedDown ) pressedUp   = false, pressedDown  = false; // conflicting keypresses.

	if(pressedLeft)							// Left Key Pressed
	{							
		_dang += 1.0f;						// Rotate ship left
		if(++_currImg > SHIP_IMG_LEFT)		// and select a more left-tilted image
			_currImg = SHIP_IMG_LEFT;
		if(_dang > 25.0f)
			_dang = 25.0f;	// limit spinning
	}
	else if(pressedRight)					// Right Key Pressed
	{					
		_dang -= 1.0f;						// Rotate ship right
		if(--_currImg < SHIP_IMG_RIGHT)		// and select a more right-tilted image
			_currImg = SHIP_IMG_RIGHT;
		if(_dang < -25.0f)
			_dang = -25.0f;	// limit spinning
	}
	else												// Otherwise, no turning keys were pressed
	{													
		ApproachValue(_dang, 0.0f, 0.125f, 0.0001f);	// Gradually stop the ship from rotating

		if(_currImg < SHIP_IMG_FLAT)					// and select a less-tilted image
			_currImg++;
		else if(_currImg > SHIP_IMG_FLAT)
			_currImg--;
	}

	if(pressedUp)										// Increase ship velocity if thrust key is pressed
		_vel += Vec2f::FromAngleDegreesCW(_ang, 0.4f);	
	else if(pressedDown)								// Decrease ship velocity if reverse-thrust key is pressed
		_vel -= Vec2f::FromAngleDegreesCW(_ang, 0.4f);	
	else												// Otherwise, assume they mean to stop, so
		_vel *= 0.9f;									// gradually (sort of) bring ship to a stop

	if(pressedFire)						// Fire Key Pressed
	{
		if(_refire > 3)					// If refire counter is ready..
		{
			// Determine direction & velocity of the new 'fired' photon
			const Vec2f fireVelocity = _vel + Vec2f::FromAngleDegreesCW(_ang, 20.0f);

			// Add a photon to the list of photons to be processed
			g_photons.AddPhoton(_pos, fireVelocity);

			// Play photon torpedo sound (at 50% volume, since it's loud)
			GameX.PlaySound(&_photonSnd, PLAY_REWIND, 0.5f);

			// Reset the refire readiness counter
			_refire = 0;
		}
	} 	
												//*** Move the ship
	_ang += _dang;								// Adjust angle by change-in-angle	
	_pos += _vel;								// Move ship position by ship's velocity
	_refire++;									// Increment the refire readiness counter

	// Wall Rebound
	if(_pos.x < 0) _vel.x *= -0.9f, _pos.x = 0;		// If ship went out of the view,
	if(_pos.y < 0) _vel.y *= -0.9f, _pos.y = 0;		// bring it back into view.
	if(_pos.x >= GameX.GetWidth())  _vel.x *= -0.9f, _pos.x = (float)GameX.GetWidth()-1;
	if(_pos.y >= GameX.GetHeight()) _vel.y *= -0.9f, _pos.y = (float)GameX.GetHeight()-1;
}

// Ship - CollideCheck
// Do ship collision detection with another ship
// (basic circle-to-circle collision detection and reaction)
void Ship::CollideCheck (Ship* ship2)
{
	assert(ship2 != this); // make sure we're not trying to collide with ourself

	const float distSq = _pos.DistSq(ship2->_pos);			// square of distance to other ship
	const float collisionRangeSq = Square(28 * 1.0f * 2);	// square of the distance at which we will collide if any closer (NOTE: 28 is the approximate radius in pixels of a single ship's image)

	if(distSq < collisionRangeSq && distSq != 0)	// If distance to 2nd ship is less than a certain number of pixels,
	{
		// Collision detection is done (now we know there was a collision)
		// Begin of collision reaction code:

		const Vec2f dir = _pos.DirectionTo(ship2->_pos) * collisionRangeSq / distSq; // Calculate direction vector to 2nd ship

		float speed = _vel.Length();			// Calculate speed of this ship
		float speed2 = ship2->_vel.Length();	// Calculate speed of other ship

		// Calculate 'rebound' amount to throw ships in opposite directions
		// (in some way that hopefully doesn't violate conservation of energy too badly)
		const float myRebound =    ((speed + speed2) * 0.5f * FactorOfSum(ship2->_weight, _weight));
		const float ship2Rebound = ((speed2 + speed) * 0.5f * FactorOfSum(_weight, ship2->_weight));

		_vel -= myRebound * dir;			// Add rebound to this ship's velocity
		ship2->_vel += ship2Rebound * dir;	// Add rebound to the other ship's velocity

		// Make the ships face a little bit toward where they were pushed
		// (just a hack; we're not correctly modeling rotational inertia or anything like that)
		ApproachAngleDegrees(_ang, _vel.ToAngleDegrees()-90, myRebound*0.0125f, 0.0f);
		ApproachAngleDegrees(ship2->_ang, ship2->_vel.ToAngleDegrees()-90, ship2Rebound*0.0125f, 0.0f);
	}
}

// Ship - Draw
// Draw the ship's shadow, target, and the ship itself
void Ship::Draw (void)
{
	int left = (_currImg % 5) * (80 + 1); // pixel x-value of sprite start in source "ships" image
	int top = (_currImg / 5) * (80 + 1); // pixel y-value of sprite start in source "ships" image

	// Draw the ship's shadow (by drawing the ship sprite totally black and halfway translucent)
	GameX.SetDrawShading(ColorX(0,0,0,128)); // (0,0,0,128) corresponds to (red,green,blue,opacity), each is out of 255
	GameX.SetDrawPart(left, top, left+80, top+80); // the individual ship sprite is 80x80 pixels
	GameX.SetDrawMode(DRAWOP_CENTERED);
	const Vec2f shadowOffset = (_pos - GameX.GetScreenCenter())*0.05f; // we offset the shadow 6 pixels away from the screen's center
	GameX.DrawImage(&_sprites, _pos + shadowOffset, _ang-90, 1.0f); 

	// Draw the ship itself, note we draw this AFTER the shadow so it shows up on top of the shadow
	GameX.SetDrawPart(left, top, left+80, top+80);
	GameX.SetDrawMode(DRAWOP_CENTERED);
	GameX.DrawImage(&_sprites, _pos, _ang-90, 1.0f); // (_ang-90) is rotation angle, 1.0 means 100% scaling

	// Draw ship's "target" image
	GameX.SetDrawMode(DRAWOP_CENTERED);
	GameX.DrawImage(&_targetImg, _pos + _vel * 4.0f + Vec2f::FromAngleDegreesCW(_ang, 100.0f));
}